# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

import pytest
from slint import _load_file_checked
from pathlib import Path
import base64
import gzip


def base_dir() -> Path:
    assert __spec__
    origin = __spec__.origin
    assert origin is not None
    base_dir = Path(origin).parent
    assert base_dir is not None
    return base_dir


def compress_and_encode(json: str) -> str:
    return base64.standard_b64encode(gzip.compress(json.encode("utf-8"))).decode(
        "utf-8"
    )


def test_no_change() -> None:
    _load_file_checked(
        base_dir() / "api-match.slint",
        expected_api_base64_compressed=compress_and_encode(r"""
        {
            "version":"1.0",
            "globals":[],
            "components":[
                {
                    "name":"Test",
                    "properties":[
                        {
                            "name": "name",
                            "ty": "str"
                        }
                    ],
                    "aliases":[]
                }
            ],
            "structs_and_enums":[]
    }"""),
        generated_file="/some/path.py",
    )


def test_incompatible_changes() -> None:
    with pytest.raises(RuntimeError) as excinfo:
        _load_file_checked(
            base_dir() / "api-match.slint",
            expected_api_base64_compressed=compress_and_encode(r"""
            {
                "version":"1.0",
                "globals":[],
                "components":[
                    {
                        "name":"Test",
                        "properties":[
                            {
                                "name": "name",
                                "ty": "str"
                            },
                            {
                                "name": "not_there_anymore",
                                "ty": "str"
                            }
                        ],
                        "aliases":[]
                    }
                ],
                "structs_and_enums":[]
        }"""),
            generated_file="/some/path.py",
        )
    assert (
        f"Incompatible API changes detected between /some/path.py and {base_dir() / 'api-match.slint'}"
        == str(excinfo.value)
    )



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

from slint import load_file
import slint
import pytest
from pathlib import Path
import asyncio


def base_dir() -> Path:
    assert __spec__
    origin = __spec__.origin
    assert origin is not None
    base_dir = Path(origin).parent
    assert base_dir is not None
    return base_dir


def test_callback_decorators(caplog: pytest.LogCaptureFixture) -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    class SubClass(module.App):  # type: ignore
        @slint.callback()
        def say_hello_again(self, arg: str) -> str:
            return "say_hello_again:" + arg

        @slint.callback(name="say-hello")
        def renamed(self, arg: str) -> str:
            return "renamed:" + arg

        @slint.callback(global_name="MyGlobal", name="global-callback")
        def global_callback(self, arg: str) -> str:
            return "global:" + arg

    instance = SubClass()
    assert instance.invoke_say_hello("ok") == "renamed:ok"
    assert instance.invoke_say_hello_again("ok") == "say_hello_again:ok"
    assert instance.invoke_global_callback("ok") == "global:ok"
    del instance


def test_callback_decorators_async() -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    class SubClass(module.App):  # type: ignore
        def __init__(self, in_queue: asyncio.Queue[int], out_queue: asyncio.Queue[int]):
            super().__init__()
            self.in_queue = in_queue
            self.out_queue = out_queue

        @slint.callback()
        async def call_void(self) -> None:
            value = await self.in_queue.get()
            await self.out_queue.put(value + 1)

        @slint.callback()
        async def call_void2(self) -> None:
            value = await self.in_queue.get()
            await self.out_queue.put(value + 2)

    async def main(
        instance: SubClass, in_queue: asyncio.Queue[int], out_queue: asyncio.Queue[int]
    ) -> None:
        await in_queue.put(42)
        instance.invoke_call_void()
        assert await out_queue.get() == 43
        await in_queue.put(43)
        instance.invoke_call_void2()
        assert await out_queue.get() == 45
        slint.quit_event_loop()

    in_queue: asyncio.Queue[int] = asyncio.Queue()
    out_queue: asyncio.Queue[int] = asyncio.Queue()
    instance = SubClass(in_queue, out_queue)
    slint.run_event_loop(main(instance, in_queue, out_queue))


def test_callback_decorators_async_err() -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    class SubClass(module.App):  # type: ignore
        def __init__(self) -> None:
            super().__init__()

        @slint.callback()
        async def say_hello(self, msg: str) -> str:
            return msg

    with pytest.raises(RuntimeError) as excinfo:
        SubClass()
    err = excinfo.value
    assert (
        str(err)
        == "Callback 'say_hello' cannot be used with a callback decorator for an async function, as it doesn't return void"
    )



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

import pytest
from slint import load_file, ListModel
from pathlib import Path


def base_dir() -> Path:
    assert __spec__
    origin = __spec__.origin
    assert origin is not None
    base_dir = Path(origin).parent
    assert base_dir is not None
    return base_dir


def test_enums() -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    TestEnum = module.TestEnum

    assert TestEnum.Variant1.name == "Variant1"
    assert TestEnum.Variant1.value == "Variant1"
    assert TestEnum.Variant2.name == "Variant2"
    assert TestEnum.Variant2.value == "Variant2"
    with pytest.raises(
        AttributeError, match="type object 'TestEnum' has no attribute 'Variant3'"
    ):
        TestEnum.Variant3

    instance = module.App()
    assert instance.enum_property == TestEnum.Variant2
    assert instance.enum_property.__class__ is TestEnum
    instance.enum_property = TestEnum.Variant1
    assert instance.enum_property == TestEnum.Variant1
    assert instance.enum_property.__class__ is TestEnum
    instance.enum_property = TestEnum.Variant_three
    assert instance.enum_property == TestEnum.Variant_three
    assert instance.enum_property.__class__ is TestEnum

    model_with_enums = instance.model_with_enums
    assert len(model_with_enums) == 2
    assert model_with_enums[0] == TestEnum.Variant2
    assert model_with_enums[1] == TestEnum.Variant1
    assert model_with_enums[0].__class__ is TestEnum
    model_with_enums = None

    instance.model_with_enums = ListModel([TestEnum.Variant1, TestEnum.Variant2])
    assert len(instance.model_with_enums) == 2
    assert instance.model_with_enums[0] == TestEnum.Variant1
    assert instance.model_with_enums[1] == TestEnum.Variant2
    assert instance.model_with_enums[0].__class__ is TestEnum



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

from slint import slint as native
from slint.slint import ValueType
from pathlib import Path


def test_basic_compiler() -> None:
    compiler = native.Compiler()

    assert compiler.include_paths == []
    compiler.include_paths = [Path("testing")]
    assert compiler.include_paths == [Path("testing")]

    assert len(compiler.build_from_source("Garbage", Path("")).component_names) == 0

    result = compiler.build_from_source(
        """
        export global TestGlobal {
            in property <string> theglobalprop;
            callback globallogic();
            public function globalfun() {}
        }

        export component Test {
            in property <string> strprop;
            in property <int> intprop;
            in property <float> floatprop;
            in property <bool> boolprop;
            in property <image> imgprop;
            in property <brush> brushprop;
            in property <color> colprop;
            in property <[string]> modelprop;

            callback test-callback();
            public function ff() {}
        }
    """,
        Path(""),
    )
    assert result.component_names == ["Test"]
    compdef = result.component("Test")

    assert compdef is not None

    assert compdef.name == "Test"

    props = [(name, type) for name, type in compdef.properties.items()]
    assert props == [
        ("boolprop", ValueType.Bool),
        ("brushprop", ValueType.Brush),
        ("colprop", ValueType.Brush),
        ("floatprop", ValueType.Number),
        ("imgprop", ValueType.Image),
        ("intprop", ValueType.Number),
        ("modelprop", ValueType.Model),
        ("strprop", ValueType.String),
    ]

    assert compdef.callbacks == ["test-callback"]
    assert compdef.functions == ["ff"]

    assert compdef.globals == ["TestGlobal"]

    assert compdef.global_properties("Garbage") is None
    assert [
        (name, type) for name, type in compdef.global_properties("TestGlobal").items()
    ] == [("theglobalprop", ValueType.String)]

    assert compdef.global_callbacks("Garbage") is None
    assert compdef.global_callbacks("TestGlobal") == ["globallogic"]

    assert compdef.global_functions("Garbage") is None
    assert compdef.global_functions("TestGlobal") == ["globalfun"]

    instance = compdef.create()
    assert instance is not None


def test_compiler_build_from_path() -> None:
    compiler = native.Compiler()

    result = compiler.build_from_path(Path("Nonexistent.slint"))
    assert len(result.component_names) == 0

    diags = result.diagnostics
    assert len(diags) == 1

    assert diags[0].level == native.DiagnosticLevel.Error
    assert diags[0].message.startswith("Could not load Nonexistent.slint:")



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

from . import slint as native
import asyncio.selector_events
import asyncio
import asyncio.events
import selectors
import typing
from collections.abc import Mapping
import datetime


class HasFileno(typing.Protocol):
    def fileno(self) -> int: ...


def fd_for_fileobj(fileobj: int | HasFileno) -> int:
    if isinstance(fileobj, int):
        return fileobj
    return int(fileobj.fileno())


class _SlintSelectorMapping(Mapping[typing.Any, selectors.SelectorKey]):
    def __init__(self, slint_selector: "_SlintSelector") -> None:
        self.slint_selector = slint_selector

    def __len__(self) -> int:
        return len(self.slint_selector.fd_to_selector_key)

    def get(self, fileobj, default=None):  # type: ignore
        fd = fd_for_fileobj(fileobj)
        return self.slint_selector.fd_to_selector_key.get(fd, default)

    def __getitem__(self, fileobj: typing.Any) -> selectors.SelectorKey:
        fd = fd_for_fileobj(fileobj)
        return self.slint_selector.fd_to_selector_key[fd]

    def __iter__(self):  # type: ignore
        return iter(self.slint_selector.fd_to_selector_key)


class _SlintSelector(selectors.BaseSelector):
    def __init__(self) -> None:
        self.fd_to_selector_key: typing.Dict[typing.Any, selectors.SelectorKey] = {}
        self.mapping = _SlintSelectorMapping(self)
        self.adapters: typing.Dict[int, native.AsyncAdapter] = {}

    def register(
        self, fileobj: typing.Any, events: typing.Any, data: typing.Any = None
    ) -> selectors.SelectorKey:
        fd = fd_for_fileobj(fileobj)
        key = selectors.SelectorKey(fileobj, fd, events, data)
        self.fd_to_selector_key[fd] = key

        adapter = native.AsyncAdapter(fd)
        self.adapters[fd] = adapter

        if events & selectors.EVENT_READ:
            adapter.wait_for_readable(self.read_notify)
        if events & selectors.EVENT_WRITE:
            adapter.wait_for_writable(self.write_notify)

        return key

    def unregister(self, fileobj: typing.Any) -> selectors.SelectorKey:
        fd = fd_for_fileobj(fileobj)
        key = self.fd_to_selector_key.pop(fd)

        try:
            del self.adapters[fd]
        except KeyError:
            pass

        return key

    def modify(
        self, fileobj: typing.Any, events: int, data: typing.Any = None
    ) -> selectors.SelectorKey:
        fd = fd_for_fileobj(fileobj)
        key = self.fd_to_selector_key[fd]

        if key.events != events:
            self.unregister(fileobj)
            key = self.register(fileobj, events, data)
        elif key.data != data:
            key._replace(data=data)
            self.fd_to_selector_key[fd] = key

        return key

    def select(
        self, timeout: float | None = None
    ) -> typing.List[typing.Tuple[selectors.SelectorKey, int]]:
        raise NotImplementedError

    def close(self) -> None:
        pass

    def get_map(self) -> Mapping[int | HasFileno, selectors.SelectorKey]:
        return self.mapping

    def read_notify(self, fd: int) -> None:
        key = self.fd_to_selector_key[fd]
        (reader, writer) = key.data
        reader._run()

    def write_notify(self, fd: int) -> None:
        key = self.fd_to_selector_key[fd]
        (reader, writer) = key.data
        writer._run()


class SlintEventLoop(asyncio.SelectorEventLoop):
    def __init__(self) -> None:
        self._is_running = False
        self._timers: typing.Set[native.Timer] = set()
        self.stop_run_forever_event = asyncio.Event()
        self._soon_tasks: typing.List[asyncio.TimerHandle] = []
        super().__init__(_SlintSelector())

    def run_forever(self) -> None:
        async def loop_stopper(event: asyncio.Event) -> None:
            await event.wait()
            native.quit_event_loop()

        asyncio.events._set_running_loop(self)
        self._is_running = True
        try:
            self.stop_run_forever_event = asyncio.Event()
            self.create_task(loop_stopper(self.stop_run_forever_event))
            native.run_event_loop()
        finally:
            self._is_running = False
            asyncio.events._set_running_loop(None)

    def run_until_complete[T](self, future: typing.Awaitable[T]) -> T | None:  # type: ignore[override]
        def stop_loop(future: typing.Any) -> None:
            self.stop()

        future = asyncio.ensure_future(future, loop=self)
        future.add_done_callback(stop_loop)

        try:
            self.run_forever()
        finally:
            future.remove_done_callback(stop_loop)

        if future.done():
            return future.result()
        else:
            if self.stop_run_forever_event.is_set():
                raise RuntimeError("run_until_complete's future isn't done", future)
            else:
                # If the loop was quit for example because the user closed the last window, then
                # don't thrown an error but return a None sentinel. The return value of asyncio.run()
                # isn't used by slint.run_event_loop() anyway
                # TODO: see if we can properly cancel the future by calling cancel() and throwing
                # the task cancellation exception.
                return None

    def _run_forever_setup(self) -> None:
        pass

    def _run_forever_cleanup(self) -> None:
        pass

    def stop(self) -> None:
        self.stop_run_forever_event.set()

    def is_running(self) -> bool:
        return self._is_running

    def close(self) -> None:
        super().close()

    def is_closed(self) -> bool:
        return False

    def call_later(self, delay, callback, *args, context=None) -> asyncio.TimerHandle:  # type: ignore
        timer = native.Timer()

        handle = asyncio.TimerHandle(
            when=self.time() + delay,
            callback=callback,
            args=args,
            loop=self,
            context=context,
        )

        timers = self._timers

        def timer_done_cb() -> None:
            timers.remove(timer)
            if not handle._cancelled:
                handle._run()

        timer.start(
            native.TimerMode.SingleShot,
            interval=datetime.timedelta(seconds=delay),
            callback=timer_done_cb,
        )

        timers.add(timer)

        return handle

    def call_at(self, when, callback, *args, context=None) -> asyncio.TimerHandle:  # type: ignore
        return self.call_later(when - self.time(), callback, *args, context=context)

    def call_soon(self, callback, *args, context=None) -> asyncio.TimerHandle:  # type: ignore
        # Collect call-soon tasks in a separate list to ensure FIFO order, as there's no guarantee
        # that multiple single-shot timers in Slint are run in order.
        handle = asyncio.TimerHandle(
            when=self.time(), callback=callback, args=args, loop=self, context=context
        )
        self._soon_tasks.append(handle)
        self.call_later(0, self._flush_soon_tasks)
        return handle

    def _flush_soon_tasks(self) -> None:
        tasks_now = self._soon_tasks
        self._soon_tasks = []
        for handle in tasks_now:
            if not handle._cancelled:
                handle._run()

    def call_soon_threadsafe(self, callback, *args, context=None) -> asyncio.Handle:  # type: ignore
        handle = asyncio.Handle(
            callback=callback,
            args=args,
            loop=self,
            context=context,
        )

        def run_handle_cb() -> None:
            if not handle._cancelled:
                handle._run()

        native.invoke_from_event_loop(run_handle_cb)
        return handle

    def _write_to_self(self) -> None:
        raise NotImplementedError



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

import slint
import numpy as np
from PIL import Image
from pathlib import Path


def base_dir() -> Path:
    assert __spec__
    origin = __spec__.origin
    assert origin is not None
    base_dir = Path(origin).parent
    assert base_dir is not None
    return base_dir


def test_image_loading() -> None:
    image = Image.open(
        base_dir() / ".." / ".." / ".." / ".." / "logo" / "slint-logo-simple-dark.png"
    )
    assert image.size == (282, 84)
    array = np.array(image)
    slint_image = slint.Image.load_from_array(array)
    assert slint_image.width == 282
    assert slint_image.height == 84



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

import slint
from slint import slint as native
import asyncio
import typing
import aiohttp
from aiohttp import web
import socket
import threading
import pytest
import sys
import platform
from datetime import timedelta


def test_async_basic() -> None:
    async def quit_soon(call_check: typing.List[bool]) -> None:
        await asyncio.sleep(1)
        call_check[0] = True
        slint.quit_event_loop()

    call_check = [False]

    slint.run_event_loop(quit_soon(call_check))

    assert call_check[0]


def test_async_aiohttp() -> None:
    def probe_port() -> int:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("127.0.0.1", 0))
        port = typing.cast(int, s.getsockname()[1])
        # This is a race condition, but should be good enough for test environments
        s.close()
        return port

    async def hello(request: web.Request) -> web.Response:
        return web.Response(text="Hello, world")

    async def run_network_requests(
        port: int, exceptions: typing.List[Exception]
    ) -> None:
        try:
            app = web.Application()
            app.add_routes([web.get("/", hello)])
            runner = web.AppRunner(app)
            await runner.setup()

            site = web.TCPSite(runner, "127.0.0.1", port)
            await site.start()

            async with aiohttp.ClientSession() as session:
                async with session.get(f"http://127.0.0.1:{port}") as response:
                    #
                    print("Status:", response.status)
                    print("Content-type:", response.headers["content-type"])
                    #
                    html = await response.text()
                    print("Body:", html[:15], "...")
                    assert html == "Hello, world"

            await runner.cleanup()
        except Exception as e:
            exceptions.append(e)
        finally:
            slint.quit_event_loop()

    exceptions: typing.List[Exception] = []
    slint.run_event_loop(run_network_requests(probe_port(), exceptions))
    assert len(exceptions) == 0


def test_basic_socket() -> None:
    def server_thread(server_socket: socket.socket) -> None:
        server_socket.listen(1)
        conn, _ = server_socket.accept()
        try:
            data = conn.recv(1024)
            if data == b"ping":
                conn.sendall(b"pong")
            else:
                conn.sendall(b"error")
        finally:
            conn.close()
            server_socket.close()

    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.bind(("127.0.0.1", 0))
    port = server_socket.getsockname()[1]
    thread = threading.Thread(target=server_thread, args=(server_socket,))
    thread.start()

    async def run_network_request(port: int) -> None:
        reader, writer = await asyncio.open_connection("127.0.0.1", port)

        writer.write(b"ping")
        await writer.drain()

        response = []
        while chunk := await reader.read(1024):
            response.append(chunk)

        writer.close()
        await writer.wait_closed()

        assert response[0] == b"pong"
        slint.quit_event_loop()

    slint.run_event_loop(run_network_request(port))
    thread.join()


def test_server_socket() -> None:
    async def handle_client(
        reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        data = await reader.read(1024)
        if data == b"ping":
            writer.write(b"pong")
        else:
            writer.write(b"error")
        await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def run_network_request(port: int) -> None:
        try:
            reader, writer = await asyncio.open_connection("127.0.0.1", port)

            writer.write(b"ping")
            await writer.drain()

            response = []
            while chunk := await reader.read(1024):
                response.append(chunk)

            writer.close()
            await writer.wait_closed()

            assert response[0] == b"pong"
        finally:
            slint.quit_event_loop()

    async def run_server_and_client(exception_check: typing.List[Exception]) -> None:
        try:
            server = await asyncio.start_server(handle_client, "127.0.0.1", 0)
            port = server.sockets[0].getsockname()[1]

            async with server:
                await asyncio.gather(
                    server.serve_forever(),
                    run_network_request(port),
                )
        except Exception as e:
            exception_check.append(e)
            raise

    exception_check: typing.List[Exception] = []
    slint.run_event_loop(run_server_and_client(exception_check))
    if len(exception_check) > 0:
        raise exception_check[0]


def test_loop_close_while_main_future_runs() -> None:
    def q() -> None:
        native.quit_event_loop()

    async def never_quit() -> None:
        loop = asyncio.get_running_loop()
        # Call native.quit_event_loop() directly as if the user closed the last window. We should gracefully
        # handle that the future that this function represents isn't terminated.
        loop.call_later(0.1, q)
        while True:
            await asyncio.sleep(1)

    try:
        slint.run_event_loop(never_quit())
    except Exception:
        pytest.fail("Should not throw a run-time error")


def test_loop_continues_when_main_coro_finished() -> None:
    async def quit_later(quit_event: asyncio.Event) -> None:
        await quit_event.wait()
        slint.quit_event_loop()

    async def simple(quit_event: asyncio.Event) -> None:
        loop = asyncio.get_event_loop()
        loop.create_task(quit_later(quit_event))

    quit_event = asyncio.Event()
    slint.Timer.single_shot(
        duration=timedelta(milliseconds=100), callback=lambda: quit_event.set()
    )
    slint.run_event_loop(simple(quit_event))
    assert quit_event.is_set()


@pytest.mark.skipif(platform.system() == "Windows", reason="pipes aren't supported yet")
def test_subprocess() -> None:
    async def launch_process(exception_check: typing.List[Exception]) -> None:
        try:
            proc = await asyncio.create_subprocess_exec(
                sys.executable,
                "--version",
                stdout=asyncio.subprocess.PIPE,  # type: ignore
                stderr=asyncio.subprocess.STDOUT,  # type: ignore
            )

            stdout, _ = await proc.communicate()
            output = stdout.decode().strip()
            print(f"Process output: {output}")

            assert proc.returncode == 0
            assert output != ""
            slint.quit_event_loop()
        except Exception as e:
            exception_check[0] = e
            raise

    exception_check: typing.List[Exception] = []
    slint.run_event_loop(launch_process(exception_check))
    if len(exception_check) > 0:
        raise exception_check[0]


def test_exception_thrown() -> None:
    async def throws() -> None:
        raise RuntimeError("Boo")

    with pytest.raises(RuntimeError, match="Boo"):
        slint.run_event_loop(throws())



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

r"""
.. include:: ../README.md
"""

import os
import sys
from . import slint as native
import types
import logging
import copy
import typing
from typing import Any
import pathlib
from .models import ListModel, Model
from .slint import Image, Color, Brush, Timer, TimerMode
from .loop import SlintEventLoop
from pathlib import Path
from collections.abc import Coroutine
import asyncio
import gettext
import gzip
import base64

Struct = native.PyStruct


class CompileError(Exception):
    message: str
    """The error message that produced this compile error."""

    diagnostics: list[native.PyDiagnostic]
    """A list of detailed diagnostics that were produced as part of the compilation."""

    def __init__(self, message: str, diagnostics: list[native.PyDiagnostic]):
        """@private"""
        super().__init__(message)
        self.message = message
        self.diagnostics = diagnostics
        for diag in self.diagnostics:
            self.add_note(str(diag))


class Component:
    """Component is the base class for all instances of Slint components. Use the member functions to show or hide the
    window, or spin the event loop."""

    __instance__: native.ComponentInstance

    def show(self) -> None:
        """Shows the window on the screen."""

        self.__instance__.show()

    def hide(self) -> None:
        """Hides the window from the screen."""

        self.__instance__.hide()

    def run(self) -> None:
        """Shows the window, runs the event loop, hides it when the loop is quit, and returns."""
        self.show()
        run_event_loop()
        self.hide()


def _normalize_prop(name: str) -> str:
    return name.replace("-", "_")


def _build_global_class(compdef: native.ComponentDefinition, global_name: str) -> Any:
    properties_and_callbacks = {}

    for prop_name in compdef.global_properties(global_name).keys():
        python_prop = _normalize_prop(prop_name)
        if python_prop in properties_and_callbacks:
            logging.warning(f"Duplicated property {prop_name}")
            continue

        def mk_setter_getter(prop_or_callback_name: str) -> property:
            def getter(self: Component) -> Any:
                return self.__instance__.get_global_property(
                    global_name, prop_or_callback_name
                )

            def setter(self: Component, value: Any) -> None:
                self.__instance__.set_global_property(
                    global_name, prop_or_callback_name, value
                )

            return property(getter, setter)

        properties_and_callbacks[python_prop] = mk_setter_getter(prop_name)

    for callback_name in compdef.global_callbacks(global_name):
        python_prop = _normalize_prop(callback_name)
        if python_prop in properties_and_callbacks:
            logging.warning(f"Duplicated property {prop_name}")
            continue

        def mk_setter_getter(prop_or_callback_name: str) -> property:
            def getter(self: Component) -> typing.Callable[..., Any]:
                def call(*args: Any) -> Any:
                    return self.__instance__.invoke_global(
                        global_name, prop_or_callback_name, *args
                    )

                return call

            def setter(self: Component, value: typing.Callable[..., Any]) -> None:
                self.__instance__.set_global_callback(
                    global_name, prop_or_callback_name, value
                )

            return property(getter, setter)

        properties_and_callbacks[python_prop] = mk_setter_getter(callback_name)

    for function_name in compdef.global_functions(global_name):
        python_prop = _normalize_prop(function_name)
        if python_prop in properties_and_callbacks:
            logging.warning(f"Duplicated function {prop_name}")
            continue

        def mk_getter(function_name: str) -> property:
            def getter(self: Component) -> typing.Callable[..., Any]:
                def call(*args: Any) -> Any:
                    return self.__instance__.invoke_global(
                        global_name, function_name, *args
                    )

                return call

            return property(getter)

        properties_and_callbacks[python_prop] = mk_getter(function_name)

    return type("SlintGlobalClassWrapper", (), properties_and_callbacks)


def _build_class(
    compdef: native.ComponentDefinition,
) -> typing.Callable[..., Component]:
    def cls_init(self: Component, **kwargs: Any) -> Any:
        self.__instance__ = compdef.create()
        for name, value in self.__class__.__dict__.items():
            if hasattr(value, "slint.callback"):
                callback_info = getattr(value, "slint.callback")
                name = callback_info["name"]

                is_async = getattr(value, "slint.async", False)
                if is_async:
                    if "global_name" in callback_info:
                        global_name = callback_info["global_name"]
                        is_void = compdef.global_callback_returns_void(
                            global_name, name
                        )
                        if is_void is None:
                            raise AttributeError(
                                f"Callback '{name}' in global '{global_name}' cannot be used with a callback decorator for an async function, as it is not declared in Slint component"
                            )
                        if not is_void:
                            raise RuntimeError(
                                f"Callback '{name}' in global '{global_name}' cannot be used with a callback decorator for an async function, as it doesn't return void"
                            )
                    else:
                        is_void = compdef.callback_returns_void(name)
                        if is_void is None:
                            raise AttributeError(
                                f"Callback '{name}' cannot be used with a callback decorator for an async function, as it is not declared in Slint component"
                            )
                        if not is_void:
                            raise RuntimeError(
                                f"Callback '{name}' cannot be used with a callback decorator for an async function, as it doesn't return void"
                            )

                def mk_callback(
                    self: Any, callback: typing.Callable[..., Any]
                ) -> typing.Callable[..., Any]:
                    def invoke(*args: Any, **kwargs: Any) -> Any:
                        return callback(self, *args, **kwargs)

                    return invoke

                if "global_name" in callback_info:
                    self.__instance__.set_global_callback(
                        callback_info["global_name"], name, mk_callback(self, value)
                    )
                else:
                    self.__instance__.set_callback(name, mk_callback(self, value))

        for prop, val in kwargs.items():
            setattr(self, prop, val)

    properties_and_callbacks: dict[Any, Any] = {"__init__": cls_init}

    for prop_name in compdef.properties.keys():
        python_prop = _normalize_prop(prop_name)
        if python_prop in properties_and_callbacks:
            logging.warning(f"Duplicated property {prop_name}")
            continue

        def mk_setter_getter(prop_or_callback_name: str) -> property:
            def getter(self: Component) -> Any:
                return self.__instance__.get_property(prop_or_callback_name)

            def setter(self: Component, value: Any) -> None:
                self.__instance__.set_property(prop_or_callback_name, value)

            return property(getter, setter)

        properties_and_callbacks[python_prop] = mk_setter_getter(prop_name)

    for callback_name in compdef.callbacks:
        python_prop = _normalize_prop(callback_name)
        if python_prop in properties_and_callbacks:
            logging.warning(f"Duplicated property {prop_name}")
            continue

        def mk_setter_getter(prop_or_callback_name: str) -> property:
            def getter(self: Component) -> typing.Callable[..., Any]:
                def call(*args: Any) -> Any:
                    return self.__instance__.invoke(prop_or_callback_name, *args)

                return call

            def setter(self: Component, value: typing.Callable[..., Any]) -> None:
                self.__instance__.set_callback(prop_or_callback_name, value)

            return property(getter, setter)

        properties_and_callbacks[python_prop] = mk_setter_getter(callback_name)

    for function_name in compdef.functions:
        python_prop = _normalize_prop(function_name)
        if python_prop in properties_and_callbacks:
            logging.warning(f"Duplicated function {prop_name}")
            continue

        def mk_getter(function_name: str) -> property:
            def getter(self: Component) -> typing.Callable[..., Any]:
                def call(*args: Any) -> Any:
                    return self.__instance__.invoke(function_name, *args)

                return call

            return property(getter)

        properties_and_callbacks[python_prop] = mk_getter(function_name)

    for global_name in compdef.globals:
        global_class = _build_global_class(compdef, global_name)

        def mk_global(global_class: typing.Callable[..., Any]) -> property:
            def global_getter(self: Component) -> Any:
                wrapper = global_class()
                setattr(wrapper, "__instance__", self.__instance__)
                return wrapper

            return property(global_getter)

        properties_and_callbacks[global_name] = mk_global(global_class)

    return type("SlintClassWrapper", (Component,), properties_and_callbacks)


def _build_struct(name: str, struct_prototype: native.PyStruct) -> type:
    def new_struct(cls: Any, *args: Any, **kwargs: Any) -> native.PyStruct:
        inst = copy.copy(struct_prototype)

        for prop, val in kwargs.items():
            setattr(inst, prop, val)

        return inst

    type_dict = {
        "__new__": new_struct,
    }

    return type(name, (), type_dict)


def _load_file(
    path: str | os.PathLike[Any] | pathlib.Path,
    quiet: bool = False,
    style: typing.Optional[str] = None,
    include_paths: typing.Optional[typing.List[os.PathLike[Any] | pathlib.Path]] = None,
    library_paths: typing.Optional[
        typing.Dict[str, os.PathLike[Any] | pathlib.Path]
    ] = None,
    translation_domain: typing.Optional[str] = None,
) -> typing.Tuple[types.SimpleNamespace, native.CompilationResult]:
    """This function is the low-level entry point into Slint for instantiating components. It loads the `.slint` file at
    the specified `path` and returns a namespace with all exported components as Python classes, as well as enums, and structs.

    * `quiet`: Set to true to prevent any warnings during compilation from being printed to stderr.
    * `style`: Specify a widget style.
    * `include_paths`: Additional include paths used to look up `.slint` files imported from other `.slint` files.
    * `library_paths`: A dictionary that maps library names to their location in the file system. This is then used to look up
       library imports, such as `import { MyButton } from "@mylibrary";`.
    * `translation_domain`: The domain to use for looking up the catalogue run-time translations. This must match the
       translation domain used when extracting translations with `slint-tr-extractor`.

    """

    compiler = native.Compiler()

    if style is not None:
        compiler.style = style
    if include_paths is not None:
        compiler.include_paths = include_paths
    if library_paths is not None:
        compiler.library_paths = library_paths
    if translation_domain is not None:
        compiler.translation_domain = translation_domain

    result = compiler.build_from_path(Path(path))

    diagnostics = result.diagnostics
    if diagnostics:
        if not quiet:
            for diag in diagnostics:
                if diag.level == native.DiagnosticLevel.Warning:
                    logging.warning(diag)

        errors = [
            diag for diag in diagnostics if diag.level == native.DiagnosticLevel.Error
        ]
        if errors:
            raise CompileError(f"Could not compile {path}", diagnostics)

    module = types.SimpleNamespace()
    for comp_name in result.component_names:
        wrapper_class = _build_class(result.component(comp_name))

        setattr(module, comp_name, wrapper_class)

    structs, enums = result.structs_and_enums

    for name, struct_prototype in structs.items():
        name = _normalize_prop(name)
        struct_wrapper = _build_struct(name, struct_prototype)
        setattr(module, name, struct_wrapper)

    for name, enum_class in enums.items():
        name = _normalize_prop(name)
        setattr(module, name, enum_class)

    for orig_name, new_name in result.named_exports:
        orig_name = _normalize_prop(orig_name)
        new_name = _normalize_prop(new_name)
        setattr(module, new_name, getattr(module, orig_name))

    return (module, result)


def load_file(
    path: str | os.PathLike[Any] | pathlib.Path,
    quiet: bool = False,
    style: typing.Optional[str] = None,
    include_paths: typing.Optional[typing.List[os.PathLike[Any] | pathlib.Path]] = None,
    library_paths: typing.Optional[
        typing.Dict[str, os.PathLike[Any] | pathlib.Path]
    ] = None,
    translation_domain: typing.Optional[str] = None,
) -> types.SimpleNamespace:
    """This function is the low-level entry point into Slint for instantiating components. It loads the `.slint` file at
    the specified `path` and returns a namespace with all exported components as Python classes, as well as enums, and structs.

    * `quiet`: Set to true to prevent any warnings during compilation from being printed to stderr.
    * `style`: Specify a widget style.
    * `include_paths`: Additional include paths used to look up `.slint` files imported from other `.slint` files.
    * `library_paths`: A dictionary that maps library names to their location in the file system. This is then used to look up
       library imports, such as `import { MyButton } from "@mylibrary";`.
    * `translation_domain`: The domain to use for looking up the catalogue run-time translations. This must match the
       translation domain used when extracting translations with `slint-tr-extractor`.

    """

    return _load_file(
        path, quiet, style, include_paths, library_paths, translation_domain
    )[0]


def _load_file_checked(
    path: str | os.PathLike[Any] | pathlib.Path,
    expected_api_base64_compressed: str,
    generated_file: str | os.PathLike[Any] | pathlib.Path,
) -> types.SimpleNamespace:
    """@private"""

    module, compilation_result = _load_file(path)

    expected_api = gzip.decompress(
        base64.standard_b64decode(expected_api_base64_compressed)
    ).decode("utf-8")

    generated_api_module = native.GeneratedAPI(path=generated_file, json=expected_api)
    actual_api_module = compilation_result.generated_api

    generated_api_module.compare_generated_vs_actual(
        generated=generated_api_module, actual=actual_api_module
    )

    return module


class SlintAutoLoader:
    def __init__(self, base_dir: Path | None = None):
        self.local_dirs: typing.List[Path] | None = None
        if base_dir:
            self.local_dirs = [base_dir]

    def __getattr__(self, name: str) -> Any:
        for path in self.local_dirs or sys.path:
            dir_candidate = Path(path) / name
            if os.path.isdir(dir_candidate):
                loader = SlintAutoLoader(dir_candidate)
                setattr(self, name, loader)
                return loader

            file_candidate = dir_candidate.with_suffix(".slint")
            if os.path.isfile(file_candidate):
                type_namespace = load_file(file_candidate)
                setattr(self, name, type_namespace)
                return type_namespace

            dir_candidate = Path(path) / name.replace("_", "-")
            file_candidate = dir_candidate.with_suffix(".slint")
            if os.path.isfile(file_candidate):
                type_namespace = load_file(file_candidate)
                setattr(self, name, type_namespace)
                return type_namespace

        return None


loader = SlintAutoLoader()
"""Use the global `loader` object to load Slint files from the file system. It exposes two stages of attributes:
1. Any lookup of an attribute in the loader tries to match a file in `sys.path` with the `.slint` extension. For example
   `loader.my_component` looks for a file `my_component.slint` in the directories in `sys.path`.
2. Any lookup in the object returned by the first stage tries to match an exported component in the loaded file, or a
   struct, or enum. For example `loader.my_component.MyComponent` looks for an *exported* component named `MyComponent`
   in the file `my_component.slint`.

**Note:** The first entry in the module search path `sys.path` is the directory that contains the input script.

Example:
```python
import slint
# Look for a file `main.slint` in the current directory,
# #load & compile it, and instantiate the exported `MainWindow` component
main_window = slint.loader.main_window.MainWindow()
main_window.show()
...
```
"""


def _callback_decorator(
    callable: typing.Callable[..., Any], info: typing.Dict[str, Any]
) -> typing.Callable[..., Any]:
    if "name" not in info:
        info["name"] = typing.cast(Any, callable).__name__
    setattr(callable, "slint.callback", info)

    try:
        import inspect

        if inspect.iscoroutinefunction(callable):

            def run_as_task(*args, **kwargs) -> None:  # type: ignore
                loop = asyncio.get_event_loop()
                loop.create_task(callable(*args, **kwargs))

            setattr(run_as_task, "slint.callback", info)
            setattr(run_as_task, "slint.async", True)
            return run_as_task
    except ImportError:
        pass

    return callable


def callback(
    global_name: typing.Callable[..., Any] | str | None = None, name: str | None = None
) -> typing.Callable[..., Any]:
    """Use the callback decorator to mark a method as a callback that can be invoked from the Slint component.

    For the decorator to work, the method must be a member of a class that is Slint component.

    Example:
    ```python
    import slint

    class AppMainWindow(slint.loader.main_window.MainWindow):

        # Automatically connected to a callback button_clicked()
        # in main_window.slint's MainWindow.
        @slint.callback()
        def button_clicked(self):
            print("Button clicked")

    ...
    ```

    If your Python method has a different name from the Slint component's callback, use the `name` parameter to specify
    the correct name. Similarly, use the `global_name` parameter to specify the name of the correct global singleton in
    the Slint component.

    **Note:** The callback decorator can also be used with async functions. They will be run as task in the asyncio event loop.
    This is only supported for callbacks that don't return any value, and requires Python >= 3.13.
    """

    if callable(global_name):
        callback = global_name
        return _callback_decorator(callback, {})
    else:
        info = {}
        if name:
            info["name"] = name
        if global_name:
            info["global_name"] = global_name
        return lambda callback: _callback_decorator(callback, info)


def set_xdg_app_id(app_id: str) -> None:
    """Sets the application id for use on Wayland or X11 with [xdg](https://specifications.freedesktop.org/desktop-entry-spec/latest/)
    compliant window managers. This id must be set before the window is shown; it only applies to Wayland or X11."""

    native.set_xdg_app_id(app_id)


quit_event = asyncio.Event()


def run_event_loop(
    main_coro: typing.Optional[Coroutine[None, None, None]] = None,
) -> None:
    """Runs the main Slint event loop. If specified, the coroutine `main_coro` is run in parallel. The event loop doesn't
    terminate when the coroutine finishes, it terminates when calling `quit_event_loop()`.

    Example:
    ```python
    import slint

    ...
    image_model: slint.ListModel[slint.Image] = slint.ListModel()
    ...

    async def main_receiver(image_model: slint.ListModel) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get("http://some.server/svg-image") as response:
                svg = await response.read()
                image = slint.Image.from_svg_data(svg)
                image_model.append(image)

    ...
    slint.run_event_loop(main_receiver(image_model))
    ```

    """

    async def run_inner() -> None:
        global quit_event
        loop = typing.cast(SlintEventLoop, asyncio.get_event_loop())

        quit_task = asyncio.ensure_future(quit_event.wait(), loop=loop)

        tasks: typing.List[asyncio.Task[typing.Any]] = [quit_task]

        main_task = None
        if main_coro:
            main_task = loop.create_task(main_coro)
            tasks.append(main_task)

        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

        if main_task is not None and main_task in done:
            main_task.result()  # propagate exception if thrown
            if quit_task in pending:
                await quit_event.wait()

    global quit_event
    quit_event = asyncio.Event()
    asyncio.run(run_inner(), debug=False, loop_factory=SlintEventLoop)


def quit_event_loop() -> None:
    """Quits the running event loop in the next event processing cycle. This will make an earlier call to `run_event_loop()`
    return."""
    global quit_event
    quit_event.set()


def init_translations(translations: typing.Optional[gettext.GNUTranslations]) -> None:
    """Installs the specified translations object to handle translations originating from the Slint code.

    Example:
    ```python
    import gettext
    import slint

    translations_dir = os.path.join(os.path.dirname(__file__), "lang")
    try:
        translations = gettext.translation("my_app", translations_dir, ["de"])
        slint.install_translations(translations)
    except OSError:
        pass
    ```
    """
    native.init_translations(translations)


__all__ = [
    "CompileError",
    "Component",
    "load_file",
    "_load_file_checked",
    "loader",
    "Image",
    "Color",
    "Brush",
    "Model",
    "ListModel",
    "Timer",
    "TimerMode",
    "set_xdg_app_id",
    "callback",
    "run_event_loop",
    "quit_event_loop",
    "init_translations",
]



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

from . import slint as native
from collections.abc import Iterable
from abc import abstractmethod
import typing
from typing import Any, cast, Iterator


class Model[T](native.PyModelBase, Iterable[T]):
    """Model is the base class for feeding dynamic data into Slint views.

    Subclass Model to implement your own models, or use `ListModel` to wrap a list.

    Models are iterable and can be used in for loops."""

    def __new__(cls, *args: Any) -> "Model[T]":
        return super().__new__(cls)

    def __init__(self) -> None:
        self.init_self(self)

    def __len__(self) -> int:
        return self.row_count()

    def __getitem__(self, index: int) -> typing.Optional[T]:
        return self.row_data(index)

    def __setitem__(self, index: int, value: T) -> None:
        self.set_row_data(index, value)

    def __iter__(self) -> Iterator[T]:
        return ModelIterator(self)

    def set_row_data(self, row: int, value: T) -> None:
        """Call this method on mutable models to change the data for the given row.
        The UI will also call this method when modifying a model's data.
        Re-implement this method in a sub-class to handle the change."""
        super().set_row_data(row, value)

    @abstractmethod
    def row_data(self, row: int) -> typing.Optional[T]:
        """Returns the data for the given row.
        Re-implement this method in a sub-class to provide the data."""
        return cast(T, super().row_data(row))

    def notify_row_changed(self, row: int) -> None:
        """Call this method from a sub-class to notify the views that a row has changed."""
        super().notify_row_changed(row)

    def notify_row_removed(self, row: int, count: int) -> None:
        """Call this method from a sub-class to notify the views that
        `count` rows have been removed starting at `row`."""
        super().notify_row_removed(row, count)

    def notify_row_added(self, row: int, count: int) -> None:
        """Call this method from a sub-class to notify the views that
        `count` rows have been added starting at `row`."""
        super().notify_row_added(row, count)


class ListModel[T](Model[T]):
    """ListModel is a `Model` that stores its data in a Python list.

    Construct a ListMode from an iterable (such as a list itself).
    Use `ListModel.append()` to add items to the model, and use the
    `del` statement to remove items.

    Any changes to the model are automatically reflected in the views
    in UI they're used with.
    """

    def __init__(self, iterable: typing.Optional[Iterable[T]] = None):
        """Constructs a new ListModel from the give iterable. All the values
        the iterable produces are stored in a list."""

        super().__init__()
        if iterable is not None:
            self.list = list(iterable)
        else:
            self.list = []

    def row_count(self) -> int:
        return len(self.list)

    def row_data(self, row: int) -> typing.Optional[T]:
        return self.list[row]

    def set_row_data(self, row: int, value: T) -> None:
        self.list[row] = value
        super().notify_row_changed(row)

    def __delitem__(self, key: int | slice) -> None:
        if isinstance(key, slice):
            start, stop, step = key.indices(len(self.list))
            del self.list[key]
            count = len(range(start, stop, step))
            super().notify_row_removed(start, count)
        else:
            del self.list[key]
            super().notify_row_removed(key, 1)

    def append(self, value: T) -> None:
        """Appends the value to the end of the list."""
        index = len(self.list)
        self.list.append(value)
        super().notify_row_added(index, 1)


class ModelIterator[T](Iterator[T]):
    def __init__(self, model: Model[T]):
        self.model = model
        self.index = 0

    def __iter__(self) -> "ModelIterator[T]":
        return self

    def __next__(self) -> T:
        if self.index >= self.model.row_count():
            raise StopIteration()
        index = self.index
        self.index += 1
        data = self.model.row_data(index)
        assert data is not None
        return data



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

from slint import slint as native
from slint import models as models
import typing
from pathlib import Path


def test_model_notify() -> None:
    compiler = native.Compiler()

    compdef = compiler.build_from_source(
        """
  export component App {
    width: 300px;
    height: 300px;

    out property<length> layout-height: layout.height;
    in-out property<[length]> fixed-height-model;

    VerticalLayout {
      alignment: start;

      layout := VerticalLayout {
        for fixed-height in fixed-height-model: Rectangle {
            background: blue;
            height: fixed-height;
        }
      }
    }

  }
    """,
        Path(""),
    ).component("App")
    assert compdef is not None

    instance = compdef.create()
    assert instance is not None

    model = models.ListModel([100, 0])

    instance.set_property("fixed-height-model", model)

    assert instance.get_property("layout-height") == 100
    model.set_row_data(1, 50)
    assert instance.get_property("layout-height") == 150
    model.append(75)
    assert instance.get_property("layout-height") == 225
    del model[1:]  # type: ignore
    assert instance.get_property("layout-height") == 100

    assert isinstance(instance.get_property("fixed-height-model"), models.ListModel)


def test_model_from_list() -> None:
    compiler = native.Compiler()

    compdef = compiler.build_from_source(
        """
  export component App {
    in-out property<[int]> data: [1, 2, 3, 4];
  }
    """,
        Path(""),
    ).component("App")
    assert compdef is not None

    instance = compdef.create()
    assert instance is not None

    model = instance.get_property("data")
    assert model.row_count() == 4
    assert model.row_data(2) == 3

    instance.set_property("data", models.ListModel([0]))
    instance.set_property("data", model)
    assert list(instance.get_property("data")) == [1, 2, 3, 4]


def test_python_model_sequence() -> None:
    model = models.ListModel([1, 2, 3, 4, 5])

    assert len(model) == 5
    assert list(model) == [1, 2, 3, 4, 5]
    model[0] = 100
    assert list(model) == [100, 2, 3, 4, 5]
    assert model[2] == 3


def test_python_model_iterable() -> None:
    def test_generator(max: int) -> typing.Iterator[int]:
        i = 0
        while i < max:
            yield i
            i += 1

    model = models.ListModel(test_generator(5))

    assert len(model) == 5
    assert list(model) == [0, 1, 2, 3, 4]


def test_rust_model_sequence() -> None:
    compiler = native.Compiler()

    compdef = compiler.build_from_source(
        """
  export component App {
    in-out property<[int]> data: [1, 2, 3, 4, 5];
  }
    """,
        Path(""),
    ).component("App")
    assert compdef is not None

    instance = compdef.create()
    assert instance is not None

    model = instance.get_property("data")

    assert len(model) == 5
    assert list(model) == [1, 2, 3, 4, 5]
    assert model[2] == 3


def test_model_writeback() -> None:
    compiler = native.Compiler()

    compdef = compiler.build_from_source(
        """
  export component App {
    width: 300px;
    height: 300px;

    in-out property<[int]> model;
    callback write-to-model(int, int);
    write-to-model(index, value) => {
        self.model[index] = value
    }

  }
    """,
        Path(""),
    ).component("App")
    assert compdef is not None

    instance = compdef.create()
    assert instance is not None

    model = models.ListModel([100, 0])

    instance.set_property("model", model)

    instance.invoke("write-to-model", 1, 42)
    assert list(instance.get_property("model")) == [100, 42]
    instance.invoke("write-to-model", 0, 25)
    assert list(instance.get_property("model")) == [25, 42]



# Copyright © Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author David Faure <david.faure@kdab.com>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

# GDB pretty printers for types defined in i_slint_core
import gdb
import gdb.printing


class SharedVectorProvider:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        try:
            size = int(self.val["inner"]["pointer"]["header"]["size"])
            return f"<SharedVector, len={size}>"
        except Exception as e:
            return f"<SharedVector: error reading size: {e}>"

    def children(self):
        try:
            size = int(self.val["inner"]["pointer"]["header"]["size"])
            if size == 0:
                return

            inner_struct = self.val["inner"]["pointer"].dereference()
            maybe_uninit = inner_struct["data"]
            elem_type = self._element_type(maybe_uninit)
            data_ptr = maybe_uninit.address.cast(elem_type.pointer())
            for i in range(size):
                yield f"[{i}]", (data_ptr + i).dereference()
        except Exception as e:
            yield "<error>", f"error reading elements: {e}"

    def display_hint(self):
        return "array"

    @staticmethod
    def _element_type(maybe_uninit_val):
        ty = maybe_uninit_val.type
        for field in ty.fields():
            if field.name == "value":
                val_type = field.type
                if val_type.code == gdb.TYPE_CODE_STRUCT:
                    for sub in val_type.fields():
                        if sub.name in ("value", "__0", "0"):
                            return sub.type
                return val_type
        raise RuntimeError(
            "unsupported MaybeUninit layout for SharedVector element type"
        )


class SharedVectorSubPrinter(gdb.printing.SubPrettyPrinter):
    def __init__(self):
        super().__init__("SharedVector")

    def __call__(self, val):
        if not self.enabled:
            return None
        t = val.type.strip_typedefs()
        if t.code == gdb.TYPE_CODE_PTR:  # also support reference to SharedVector
            try:
                val = val.dereference()
                t = val.type.strip_typedefs()
            except gdb.error:
                return None
        if (
            t.code == gdb.TYPE_CODE_STRUCT
            and t.tag
            and t.tag.startswith("i_slint_core::sharedvector::SharedVector<")
        ):
            return SharedVectorProvider(val)
        return None


class SliceProvider:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        try:
            return f"<Slice, len={int(self.val['len'])}>"
        except Exception as e:
            return f"<Slice: error reading len: {e}>"

    def children(self):
        try:
            length = int(self.val["len"])
            if length == 0:
                return

            data_ptr = self._data_pointer()
            elem_type = data_ptr.type.target()
            for i in range(length):
                yield f"[{i}]", (data_ptr + i).dereference()
        except Exception as e:
            yield "<error>", f"error reading elements: {e}"

    def display_hint(self):
        return "array"

    def _data_pointer(self):
        nn = self.val["ptr"]
        if nn.type.code == gdb.TYPE_CODE_PTR:
            return nn
        for field in nn.type.fields():
            if field.name in ("pointer", "__0", "0"):
                candidate = nn[field]
                if candidate.type.code == gdb.TYPE_CODE_PTR:
                    return candidate
        raise RuntimeError("unsupported NonNull layout in Slice")


class SliceSubPrinter(gdb.printing.SubPrettyPrinter):
    def __init__(self):
        super().__init__("Slice")

    def __call__(self, val):
        if not self.enabled:
            return None
        t = val.type.strip_typedefs()
        if t.code == gdb.TYPE_CODE_PTR:  # also support reference to Slice
            try:
                val = val.dereference()
                t = val.type.strip_typedefs()
            except gdb.error:
                return None
        if (
            t.code == gdb.TYPE_CODE_STRUCT
            and t.tag
            and t.tag.startswith("i_slint_core::slice::Slice<")
        ):
            return SliceProvider(val)
        return None


class SlintCorePrettyPrinter(gdb.printing.PrettyPrinter):
    def __init__(self):
        super().__init__("i_slint_core", [])
        self.subprinters = [SharedVectorSubPrinter(), SliceSubPrinter()]

    def __call__(self, val):
        for sp in self.subprinters:
            pp = sp(val)
            if pp is not None:
                return pp
        return None


printer = SlintCorePrettyPrinter()


def register_printers(objfile=None):
    gdb.printing.register_pretty_printer(objfile, printer, replace=True)


register_printers()



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

from slint import load_file, init_translations
from pathlib import Path
import gettext
import typing


def base_dir() -> Path:
    assert __spec__
    origin = __spec__.origin
    assert origin is not None
    base_dir = Path(origin).parent
    assert base_dir is not None
    return base_dir


class DummyTranslation:
    def gettext(self, message: str) -> str:
        if message == "Yes":
            return "Ja"
        return message

    def pgettext(self, context: str, message: str) -> str:
        return self.gettext(message)


def test_load_file() -> None:
    module = load_file(base_dir() / "test-load-file.slint")

    testcase = module.App()

    assert testcase.translated == "Yes"
    init_translations(typing.cast(gettext.GNUTranslations, DummyTranslation()))
    try:
        assert testcase.translated == "Ja"
    finally:
        init_translations(None)
        assert testcase.translated == "Yes"



# Copyright © SixtyFPS GmbH <info@slint.dev>
# SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-Royalty-free-2.0 OR LicenseRef-Slint-Software-3.0

import pytest
from slint import load_file, CompileError
from pathlib import Path


def base_dir() -> Path:
    assert __spec__
    origin = __spec__.origin
    assert origin is not None
    base_dir = Path(origin).parent
    assert base_dir is not None
    return base_dir


def test_load_file(caplog: pytest.LogCaptureFixture) -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    assert (
        "The property 'color' has been deprecated. Please use 'background' instead"
        in caplog.text
    )

    assert len(list(module.__dict__.keys())) == 7
    assert "App" in module.__dict__
    assert "Diag" in module.__dict__
    assert "MyDiag" in module.__dict__
    assert "MyData" in module.__dict__
    assert "Secret_Struct" in module.__dict__
    assert "Public_Struct" in module.__dict__
    assert "TestEnum" in module.__dict__
    instance = module.App()
    del instance
    instance = module.MyDiag()
    del instance

    struct_instance = module.MyData()
    struct_instance.name = "Test"
    struct_instance.age = 42

    struct_instance = module.MyData(name="testing")
    assert struct_instance.name == "testing"

    assert module.Public_Struct is module.Secret_Struct
    assert module.MyDiag is module.Diag


def test_load_file_fail() -> None:
    with pytest.raises(CompileError, match="Could not compile non-existent.slint"):
        load_file("non-existent.slint")


def test_compile_error() -> None:
    with pytest.raises(CompileError) as excinfo:
        load_file(base_dir() / "test-file-error.slint")
    err = excinfo.value
    diagnostics = err.diagnostics
    assert len(diagnostics) == 2
    assert diagnostics[0].message == "Unknown type 'garbage'"
    assert diagnostics[0].line_number == 6
    assert diagnostics[0].column_number == 15
    assert diagnostics[1].message == "Unknown element 'NewType'"
    assert diagnostics[1].line_number == 7
    assert diagnostics[1].column_number == 5
    path = base_dir() / "test-file-error.slint"
    assert str(err) == f"Could not compile {path}"
    assert len(err.__notes__) == 2
    assert err.__notes__[0] == f"{path}:6: Unknown type 'garbage'"
    assert err.__notes__[1] == f"{path}:7: Unknown element 'NewType'"


def test_load_file_wrapper() -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    instance = module.App()

    assert instance.hello == "World"
    instance.hello = "Ok"
    assert instance.hello == "Ok"

    instance.say_hello = lambda x: "from here: " + x
    assert instance.say_hello("wohoo") == "from here: wohoo"

    assert instance.plus_one(42) == 43

    assert instance.MyGlobal.global_prop == "This is global"

    assert instance.MyGlobal.minus_one(100) == 99

    assert instance.SecondGlobal.second == "second"

    del instance


def test_constructor_kwargs() -> None:
    module = load_file(base_dir() / "test-load-file.slint", quiet=False)

    def early_say_hello(arg: str) -> str:
        return "early:" + arg

    instance = module.App(hello="Set early", say_hello=early_say_hello)

    assert instance.hello == "Set early"
    assert instance.invoke_say_hello("test") == "early:test"

    del instance
